import re


class ResourceMatcher(object):

    def __init__(self, resources):
        self.resources = resources
        if resources is None:
            self.resources = None
        elif isinstance(self.resources, str):
            self.resources = re.compile('^' + self.resources + '$')
            self.re = True
        else:
            assert isinstance(self.resources, list)
            self.re = False

    def match(self, name):
        if self.resources is None:
            return True
        if self.re:
            return self.resources.match(name) is not None
        else:
            return name in self.resources



import re
import logging

from datapackage_pipelines.wrapper import ingest, spew
from datapackage_pipelines.utilities.resource_matcher import ResourceMatcher
import tableschema

parameters, datapackage, resource_iterator = ingest()

resources = ResourceMatcher(parameters.get('resources'))
types = parameters.get('types', {})


def match_fields(field_name_re, expected):
    def filt(field):
        return (field_name_re.fullmatch(field['name']) is not None) is expected
    return filt


def process_datapackage(datapackage_):
    for resource in datapackage_['resources']:
        name = resource['name']
        if not resources.match(name):
            continue

        if 'schema' not in resource:
            continue

        fields = resource.setdefault('schema', {}).get('fields', [])
        for field_name, field_definition in types.items():
            field_name_re = re.compile(field_name)
            if field_definition is not None:
                filtered_fields = list(
                    filter(match_fields(field_name_re, True), fields)
                )
                for field in filtered_fields:
                    field.update(field_definition)
                assert len(filtered_fields) > 0, \
                    "No field found matching %r" % field_name
            else:
                fields = list(
                    filter(match_fields(field_name_re, False), fields)
                )

        resource['schema']['fields'] = fields


def process_resource(spec, rows):
    schema = spec['schema']
    jts = tableschema.Schema(schema)
    field_names = list(map(lambda f: f['name'], schema['fields']))
    for row in rows:
        flattened_row = [row.get(name) for name in field_names]
        try:
            flattened_row = jts.cast_row(flattened_row)
        except Exception:
            logging.error('Failed to cast row %r', flattened_row)
            raise
        row = dict(zip(field_names, flattened_row))
        yield row


def process_resources(resource_iterator_):
    for resource in resource_iterator_:
        spec = resource.spec
        if not resources.match(spec['name']):
            yield resource
        else:
            yield process_resource(spec, resource)


process_datapackage(datapackage)

spew(datapackage, process_resources(resource_iterator))



import logging
import os

from ..utilities.execution_id import gen_execution_id

from ..status import status

from .dependency_manager import DependencyManager
from .celery_common import get_celery_app
from ..specs import pipelines, PipelineSpec, register_all_pipelines
from ..manager.tasks import execute_pipeline


celery_app = get_celery_app()


executed_hashes = {}
# dependencies = {}
# dependents = {}
already_init = False
dep_mgr = None


def get_dep_mgr() -> DependencyManager:
    global dep_mgr
    if dep_mgr is None:
        dep_mgr = DependencyManager()
    return dep_mgr


def build_dependents():
    dm = get_dep_mgr()
    for spec in pipelines():  # type: PipelineSpec
        dm.update(spec)


def collect_dependencies(pipeline_ids):
    if pipeline_ids is None:
        return None
    dm = get_dep_mgr()
    ret = set()
    for pipeline_id in pipeline_ids:
        ret.add(pipeline_id)
        deps = dm.get_dependencies(pipeline_id)
        if deps is not None:
            ret.update(collect_dependencies(deps))
    return ret


def queue_pipeline(spec: PipelineSpec, trigger):
    ps = status.get(spec.pipeline_id)
    ps.init(spec.pipeline_details,
            spec.source_details,
            spec.validation_errors,
            spec.cache_hash)
    if ps.runnable():
        eid = gen_execution_id()
        logging.info('%s QUEUEING %s task %s', eid[:8], trigger.upper(), spec.pipeline_id)
        if ps.queue_execution(eid, trigger):
            execute_pipeline_task.delay(spec.pipeline_id,
                                        spec.pipeline_details,
                                        spec.path,
                                        trigger,
                                        eid)
            return True
    else:
        logging.warning('Skipping %s task %s, as it has errors %r',
                        trigger.upper(), spec.pipeline_id, spec.validation_errors)
    return False


def execute_update_pipelines():
    update_pipelines.delay('update', None, None)


@celery_app.task
def update_pipelines(action, completed_pipeline_id, completed_trigger):
    # action=init: register all pipelines, trigger anything that's dirty
    # action=update: iterate over all pipelines, register new ones, trigger dirty ones
    # action=complete: iterate over all pipelines, trigger dependencies
    # completed_pipeline_id: pipeline id that had just completed (when applicable)
    # completed_trigger: the trigger for the pipeline that had just completed (when applicable)
    global already_init
    if action == 'init':
        if already_init:
            return
        else:
            register_all_pipelines()
    already_init = True

    logging.info("Update Pipelines (%s)", action)
    status_all_pipeline_ids = set(status.all_pipeline_ids())
    executed_count = 0
    all_pipeline_ids = set()
    dm = get_dep_mgr()

    if action == 'complete':
        filter = collect_dependencies(dm.get_dependents(completed_pipeline_id))
        logging.info("DEPENDENTS Pipeline: %s <- %s", completed_pipeline_id, filter)
    else:
        filter = ('',)

    for spec in pipelines(filter):  # type: PipelineSpec
        all_pipeline_ids.add(spec.pipeline_id)
        ps = status.get(spec.pipeline_id)

        if action == 'init':
            ps.init(spec.pipeline_details,
                    spec.source_details,
                    spec.validation_errors,
                    spec.cache_hash)

        elif action == 'update':
            if spec.pipeline_id not in status_all_pipeline_ids:
                ps.init(spec.pipeline_details,
                        spec.source_details,
                        spec.validation_errors,
                        spec.cache_hash)
                dm.update(spec)
                logging.info("NEW Pipeline: %s", spec)
            logging.debug('Pipeline: %s (dirty: %s, %s != %s?)',
                          spec.pipeline_id, ps.dirty(), executed_hashes.get(spec.pipeline_id), spec.cache_hash)

        elif action == 'complete':
            if completed_pipeline_id in spec.dependencies:
                ps.init(spec.pipeline_details,
                        spec.source_details,
                        spec.validation_errors,
                        spec.cache_hash)
                logging.info("DEPENDENT Pipeline: %s (%d errors) (from ...%s), trigger=%s",
                             spec.pipeline_id, len(spec.validation_errors),
                             os.path.basename(completed_pipeline_id),
                             completed_trigger)
            else:
                continue

        psle = ps.get_last_execution()
        last_successful = psle.success is True if psle is not None else False
        if ps.runnable() and \
                (ps.dirty() or
                 (completed_trigger == 'scheduled') or
                 (action == 'init' and not last_successful)):
            queued = queue_pipeline(spec, 'dirty-task-%s' % action if completed_trigger is None else completed_trigger)
            if queued:
                executed_count += 1
                if executed_count == 4 and action == 'update':
                    # Limit ops on update only
                    break

    if executed_count == 0 and action != 'complete':
        extra_pipelines = status_all_pipeline_ids.difference(all_pipeline_ids)
        for pipeline_id in extra_pipelines:
            logging.info("Removing Pipeline: %s", pipeline_id)
            status.deregister(pipeline_id)
            dm.remove(pipeline_id)


@celery_app.task
def execute_scheduled_pipeline(pipeline_id):
    for spec in pipelines([pipeline_id]):
        if spec.pipeline_id == pipeline_id:
            logging.info('Running scheduled pipeline %s, with schedule %s (%r)',
                         pipeline_id, spec.schedule,
                         spec.pipeline_details.get('schedule'))
            queue_pipeline(spec, 'scheduled')


@celery_app.task
def execute_pipeline_task(pipeline_id,
                          pipeline_details,
                          pipeline_cwd,
                          trigger,
                          execution_id):

    spec = PipelineSpec(pipeline_id=pipeline_id,
                        pipeline_details=pipeline_details,
                        path=pipeline_cwd)
    success, _, _ = \
        execute_pipeline(spec,
                         execution_id,
                         trigger,
                         False)

    if success:
        has_dependents = len(get_dep_mgr().get_dependents(spec.pipeline_id)) > 0
        logging.info('HAS DEPENDENTS %s: %s', pipeline_id, has_dependents)
        if has_dependents:
            update_pipelines.delay('complete', pipeline_id, trigger)



import itertools

from datapackage_pipelines.wrapper import ingest, spew
from datapackage_pipelines.utilities.resource_matcher import ResourceMatcher
from datapackage_pipelines.utilities.resources import PROP_STREAMING

parameters, datapackage, resource_iterator = ingest()

sources = ResourceMatcher(parameters.get('sources'))

target = parameters.get('target', {})
if 'name' not in target:
    target['name'] = 'concat'
if 'path' not in target:
    target['path'] = 'data/' + target['name'] + '.csv'
target.update(dict(
    mediatype='text/csv',
    schema=dict(fields=[], primaryKey=[]),
))
target[PROP_STREAMING] = True

fields = parameters['fields']

# Create mapping between source field names to target field names
field_mapping = {}
for target_field, source_fields in fields.items():
    if source_fields is not None:
        for source_field in source_fields:
            if source_field in field_mapping:
                raise RuntimeError('Duplicate appearance of %s (%r)' % (source_field, field_mapping))
            field_mapping[source_field] = target_field

    if target_field in field_mapping:
        raise RuntimeError('Duplicate appearance of %s' % target_field)

    field_mapping[target_field] = target_field

# Create the schema for the target resource
needed_fields = sorted(fields.keys())
for resource in datapackage['resources']:
    if not sources.match(resource['name']):
        continue

    schema = resource.get('schema', {})
    pk = schema.get('primaryKey', [])
    for field in schema.get('fields', []):
        orig_name = field['name']
        if orig_name in field_mapping:
            name = field_mapping[orig_name]
            if name not in needed_fields:
                continue
            if orig_name in pk:
                target['schema']['primaryKey'].append(name)
            target['schema']['fields'].append(field)
            field['name'] = name
            needed_fields.remove(name)

if len(target['schema']['primaryKey']) == 0:
    del target['schema']['primaryKey']

for name in needed_fields:
    target['schema']['fields'].append(dict(
        name=name, type='string'
    ))

# Update resources in datapackage (make sure they are consecutive)
prefix = True
suffix = False
num_concatenated = 0
new_resources = []
for resource in datapackage['resources']:
    name = resource['name']
    match = sources.match(name)
    if prefix:
        if match:
            prefix = False
            num_concatenated += 1
        else:
            new_resources.append(resource)
    elif suffix:
        assert not match
        new_resources.append(resource)
    else:
        if not match:
            suffix = True
            new_resources.append(target)
            new_resources.append(resource)
        else:
            num_concatenated += 1
if not suffix:
    new_resources.append(target)


datapackage['resources'] = new_resources

all_target_fields = set(fields.keys())


def concatenator(resources):
    for resource_ in resources:
        for row in resource_:
            processed = dict((k, '') for k in all_target_fields)
            values = [(field_mapping[k], v) for (k, v)
                      in row.items()
                      if k in field_mapping]
            assert len(values) > 0
            processed.update(dict(values))
            yield processed


def new_resource_iterator(resource_iterator_):
    while True:
        resource_ = next(resource_iterator_)
        if sources.match(resource_.spec['name']):
            resource_chain = \
                itertools.chain([resource_],
                                itertools.islice(resource_iterator_,
                                                 num_concatenated-1))
            yield concatenator(resource_chain)
        else:
            yield resource_


spew(datapackage, new_resource_iterator(resource_iterator))



from datapackage_pipelines.wrapper import ingest, spew
from datapackage_pipelines.utilities.kvstore import KVStore
from datapackage_pipelines.utilities.resource_matcher import ResourceMatcher


class KeyCalc(object):

    def __init__(self, key_spec):
        self.key_spec = key_spec

    def __call__(self, row):
        return self.key_spec.format(**row)


parameters, datapackage, resource_iterator = ingest()

resources = ResourceMatcher(parameters['resources'])
key_calc = KeyCalc(parameters['sort-by'])


def sorter(resource):
    db = KVStore()
    for row in resource:
        key = key_calc(row)
        db[key] = row
    for key in db.keys():
        yield db[key]


def new_resource_iterator(resource_iterator_):
    for resource in resource_iterator_:
        if resources.match(resource.spec['name']):
            yield sorter(resource)
        else:
            yield resource


spew(datapackage, new_resource_iterator(resource_iterator))



import cachetools
import tempfile
import logging

from .extended_json import json

try:
    import plyvel as DB_ENGINE
    logging.info('Using leveldb for joining')
    db_kind = 'LevelDB'
except ImportError:
    import sqlite3 as DB_ENGINE
    logging.info('Using sqlite for joining')
    db_kind = 'sqlite'


class SqliteDB(object):

    def __init__(self):
        self.tmpfile = tempfile.NamedTemporaryFile()
        self.db = DB_ENGINE.connect(self.tmpfile.name)
        self.cursor = self.db.cursor()
        self.cursor.execute('''CREATE TABLE d (key text, value text)''')
        self.cursor.execute('''CREATE UNIQUE INDEX i ON d (key)''')

    def get(self, key):
        ret = self.cursor.execute('''SELECT value FROM d WHERE key=?''',
                                  (key,)).fetchone()
        if ret is None:
            raise KeyError()
        else:
            return json.loads(ret[0])

    def set(self, key, value):
        value = json.dumps(value)
        try:
            self.get(key)
            self.cursor.execute('''UPDATE d SET value=? WHERE key=?''',
                                (value, key))
        except KeyError:
            self.cursor.execute('''INSERT INTO d VALUES (?, ?)''',
                                (key, value))
        self.db.commit()

    def keys(self):
        cursor = self.db.cursor()
        keys = cursor.execute('''SELECT key FROM d ORDER BY key ASC''')
        for key, in keys:
            yield key


class LevelDB(object):

    def __init__(self):
        self.tmpdir = tempfile.TemporaryDirectory()
        self.db = DB_ENGINE.DB(self.tmpdir.name, create_if_missing=True)

    def get(self, key):
        ret = self.db.get(key.encode('utf8'))
        if ret is None:
            raise KeyError()
        else:
            return json.loads(ret.decode('utf8'))

    def set(self, key, value):
        value = json.dumps(value).encode('utf8')
        key = key.encode('utf8')
        self.db.put(key, value)

    def keys(self):
        for key, value in self.db:
            yield key.decode('utf8')


DB = LevelDB if db_kind == 'LevelDB' else SqliteDB


class CachedDB(cachetools.LRUCache):

    def __init__(self):
        super(CachedDB, self).__init__(1024, self._dbget)
        self.db = DB()

    def popitem(self):
        key, value = super(CachedDB, self).popitem()
        self._dbset(key, value)
        return key, value

    def _dbget(self, key):
        value = self.db.get(key)
        return value

    def _dbset(self, key, value):
        assert value is not None
        self.db.set(key, value)

    def sync(self):
        for key in iter(self):
            value = cachetools.Cache.__getitem__(self, key)
            self._dbset(key, value)

    def keys(self):
        self.sync()
        return self.db.keys()


KVStore = CachedDB



import collections

from datapackage_pipelines.wrapper import ingest, spew

params, dp, res_iter = ingest()


def sink(res_iter_):
    for res in res_iter_:
        collections.deque(res, maxlen=0)
    yield from ()


spew({'name': 'boop', 'resources': []},
     sink(res_iter))



import datapackage

from datapackage_pipelines.wrapper import ingest, spew

parameters, dp, res_iter = ingest()

url = parameters['url']

datapackage = datapackage.DataPackage(url)
for k, v in datapackage.descriptor.items():
    if k != 'resources':
        dp[k] = v

spew(dp, res_iter)



import datetime
import decimal
import os
import logging

import copy
from datapackage_pipelines.utilities.extended_json import json
from tableschema_sql import Storage
from sqlalchemy import create_engine
from sqlalchemy.exc import OperationalError

from datapackage_pipelines.lib.dump.dumper_base import DumperBase
from tableschema.exceptions import ValidationError


def jsonize(obj):
    return json.dumps(obj)


def strize(obj):
    if isinstance(obj, dict):
        return dict(
            (k, strize(v))
            for k, v in obj.items()
        )
    elif isinstance(obj, (str, int, float, bool)):
        return obj
    elif isinstance(obj, datetime.date):
        return obj.isoformat()
    elif isinstance(obj, decimal.Decimal):
        return float(obj)
    elif isinstance(obj, (list, set)):
        return [strize(x) for x in obj]
    elif obj is None:
        return None
    assert False, "Don't know how to handle object %r" % obj


OBJECT_FIXERS = {
    'sqlite': [strize, jsonize],
    'postgresql': [strize]
}


class SQLDumper(DumperBase):

    def initialize(self, parameters):
        super(SQLDumper, self).initialize(parameters)
        table_to_resource = parameters['tables']
        engine = parameters.get('engine', 'env://DPP_DB_ENGINE')
        if engine.startswith('env://'):
            env_var = engine[6:]
            engine = os.environ.get(env_var)
            assert engine is not None, \
                "Couldn't connect to DB - " \
                "Please set your '%s' environment variable" % env_var
        self.engine = create_engine(engine)
        try:
            self.engine.connect()
        except OperationalError:
            logging.exception('Failed to connect to database %s', engine)
            raise

        for k, v in table_to_resource.items():
            v['table-name'] = k

        self.converted_resources = \
            dict((v['resource-name'], v) for v in table_to_resource.values())

        self.updated_column = parameters.get("updated_column")
        self.updated_id_column = parameters.get("updated_id_column")

    def handle_resource(self, resource, spec, parameters, datapackage):
        resource_name = spec['name']
        if resource_name not in self.converted_resources:
            return resource
        else:
            converted_resource = self.converted_resources[resource_name]
            mode = converted_resource.get('mode', 'rewrite')
            table_name = converted_resource['table-name']
            storage = Storage(self.engine, prefix=table_name)
            if mode == 'rewrite' and '' in storage.buckets:
                storage.delete('')
            schema = self.normalise_schema_for_engine(self.engine.dialect.name,
                                                      spec['schema'])
            if '' not in storage.buckets:
                logging.info('Creating DB table %s', table_name)
                try:
                    storage.create('', schema)
                except ValidationError as e:
                    logging.error('Error validating schema %r', spec['schema'])
                    for err in e.errors:
                        logging.error('Error validating schema: %s', err)
                    raise
            else:
                storage.describe('', schema)

            update_keys = None
            if mode == 'update':
                update_keys = converted_resource.get('update_keys')
                if update_keys is None:
                    update_keys = spec['schema'].get('primaryKey', [])
            logging.info('Writing to DB %s -> %s (mode=%s, keys=%s)',
                         resource_name, table_name, mode, update_keys)
            return map(self.get_output_row,
                       storage.write(
                           '',
                           self.normalise_for_engine(self.engine.dialect.name,
                                                     resource, spec),
                           keyed=True, as_generator=True,
                           update_keys=update_keys
                       ))

    def get_output_row(self, written):
        row, updated, updated_id = written.row, written.updated, written.updated_id
        if self.updated_column:
            row[self.updated_column] = updated
        if self.updated_id_column:
            row[self.updated_id_column] = updated_id
        return row

    def normalise_schema_for_engine(self, dialect, schema):
        schema = copy.deepcopy(schema)
        for field in schema['fields']:
            if dialect == 'sqlite' and field['type'] in ['object', 'array']:
                field['type'] = 'string'
        return schema

    def normalise_for_engine(self, dialect, resource, spec):
        actions = {}
        for field in spec['schema']['fields']:
            if field['type'] in ['array', 'object']:
                assert dialect in OBJECT_FIXERS, "Don't know how to handle %r connection dialect" % dialect
                actions.setdefault(field['name'], []).extend(OBJECT_FIXERS[dialect])

        for row in resource:
            for name, action_list in actions.items():
                for action in action_list:
                    row[name] = action(row.get(name))

            yield row


SQLDumper()()



import click

from .utilities.stat_utils import user_facing_stats
from .utilities.execution_id import gen_execution_id

from .manager.logging_config import logging

from .specs import pipelines, register_all_pipelines, PipelineSpec #noqa
from .status import status
from .manager import execute_pipeline, finalize


@click.group(invoke_without_command=True)
@click.pass_context
def cli(ctx):
    if ctx.invoked_subcommand is None:
        click.echo('Available Pipelines:')
        for spec in pipelines():  # type: PipelineSpec
            ps = status.get(spec.pipeline_id)
            click.echo('- {} {}{}'
                       .format(spec.pipeline_id,
                               '(*)' if ps.dirty() else '',
                               '(E)' if len(spec.validation_errors) > 0 else ''))
            for error in spec.validation_errors:
                click.echo('\t{}: {}'.format(error.short_msg,
                                             error.long_msg))


@cli.command()
def serve():
    """Start the web server"""
    from .web import app
    app.run(host='0.0.0.0', debug=True, port=5000)


def match_pipeline_id(arg, pipeline_id):
    if arg.endswith('*'):
        return pipeline_id.startswith(arg[:-1])
    else:
        return pipeline_id == arg


def execute_if_needed(argument, spec, use_cache):
    ps = status.get(spec.pipeline_id)
    if (match_pipeline_id(argument, spec.pipeline_id) or
            (argument == 'all') or
            (argument == 'dirty' and ps.dirty())):
        if len(spec.validation_errors) != 0:
            return (spec.pipeline_id, False, {}, ['init'] + list(map(str, spec.validation_errors))), \
                   spec.pipeline_id == argument
        eid = gen_execution_id()
        ps.queue_execution(eid, 'manual')
        success, stats, errors = \
            execute_pipeline(spec, eid,
                             use_cache=use_cache)
        return (spec.pipeline_id, success, stats, errors), spec.pipeline_id == argument
    return None, False


@cli.command()
@click.argument('pipeline_id')
@click.option('--use-cache/--no-use-cache', default=True)
def run(pipeline_id, use_cache):
    """Run a pipeline by pipeline-id.
Use 'all' for running all pipelines,
or 'dirty' for running just the dirty ones."""
    register_all_pipelines()
    try:
        results = []
        executed = set()
        modified = 1
        while modified > 0:
            modified = 0
            for spec in pipelines():

                if spec.pipeline_id in executed:
                    continue

                ret, stop = \
                    execute_if_needed(pipeline_id, spec, use_cache)

                if ret is not None:
                    executed.add(spec.pipeline_id)
                    modified += 1
                    results.append(ret)

                if stop:
                    modified = 0
                    break

        logging.info('RESULTS:')
        for pipeline_id, success, stats, errors in results:
            stats = user_facing_stats(stats)
            logging.info('%s: %s %s%s',
                         'SUCCESS' if success else 'FAILURE',
                         pipeline_id,
                         repr(stats) if stats is not None else '',
                         ('\nERROR log from processor %s:\n+--------\n| ' % errors[0] +
                          '\n| '.join(errors[1:]) +
                          '\n+--------')
                         if errors else ''
                         )

    except KeyboardInterrupt:
        pass
    finally:
        finalize()


@cli.command()
def init():
    """Reset the status of all pipelines"""
    status.initialize()


if __name__ == "__main__":
    cli()



import itertools

import datapackage

from datapackage_pipelines.wrapper import ingest, spew, get_dependency_datapackage_url
from datapackage_pipelines.utilities.resource_matcher import ResourceMatcher
from datapackage_pipelines.utilities.resources import tabular, PROP_STREAMING, \
    PROP_STREAMED_FROM


class ResourceLoader(object):

    def __init__(self):
        self.parameters, self.dp, self.res_iter = ingest()

    def __call__(self):
        url = self.parameters['url']
        dep_prefix = 'dependency://'
        if url.startswith(dep_prefix):
            dependency = url[len(dep_prefix):].strip()
            url = get_dependency_datapackage_url(dependency)
            assert url is not None, "Failed to fetch output datapackage for dependency '%s'" % dependency
        resource = self.parameters['resource']
        stream = self.parameters.get('stream', True)
        name_matcher = ResourceMatcher(resource) if isinstance(resource, str) else None
        resource_index = resource if isinstance(resource, int) else None

        selected_resources = []
        found = False
        dp = datapackage.DataPackage(url)
        dp = self.process_datapackage(dp)
        for i, orig_res in enumerate(dp.resources):
            if resource_index == i or \
                    (name_matcher is not None and name_matcher.match(orig_res.descriptor.get('name'))):
                found = True
                orig_res.descriptor[PROP_STREAMED_FROM] = orig_res.source
                self.dp['resources'].append(orig_res.descriptor)
                if tabular(orig_res.descriptor) and stream:
                    orig_res.descriptor[PROP_STREAMING] = True
                    selected_resources.append(orig_res.iter(keyed=True))
                else:
                    orig_res.descriptor[PROP_STREAMING] = False

        assert found, "Failed to find resource with index or name matching %r" % resource
        spew(self.dp, itertools.chain(self.res_iter, selected_resources))

    def process_datapackage(self, dp_):
        return dp_


if __name__ == '__main__':
    ResourceLoader()()



from datapackage_pipelines.wrapper import ingest, spew
from datapackage_pipelines.utilities.resource_matcher import ResourceMatcher

import operator

parameters, datapackage, resource_iterator = ingest()

resources = ResourceMatcher(parameters.get('resources'))
equals = parameters.get('in', [])
not_equals = parameters.get('out', [])

conditions = [
    (operator.eq, k, v)
    for o in equals
    for k, v in o.items()
] + [
    (operator.ne, k, v)
    for o in not_equals
    for k, v in o.items()
]


def process_resource(rows):
    for row in rows:
        if any(func(row[k], v) for func, k, v in conditions):
            yield row


def process_resources(resource_iterator_):
    for resource in resource_iterator_:
        spec = resource.spec
        if not resources.match(spec['name']):
            yield resource
        else:
            yield process_resource(resource)


spew(datapackage, process_resources(resource_iterator))



from datapackage_pipelines.wrapper import ingest, spew

parameters, datapackage, res_iter = ingest()
if datapackage is None:
    datapackage = parameters
else:
    datapackage.update(parameters)

spew(datapackage, res_iter)



import datetime
import os

import logging
import slugify
import yaml
import mistune
from copy import deepcopy
from flask import Blueprint

from flask import Flask, render_template, abort, redirect
from flask_cors import CORS
from flask_jsonpify import jsonify

from datapackage_pipelines.celery_tasks.celery_tasks import execute_update_pipelines
from datapackage_pipelines.status import status
from datapackage_pipelines.specs import register_all_pipelines
from datapackage_pipelines.utilities.stat_utils import user_facing_stats

register_all_pipelines()


def datestr(x):
    if x is None:
        return ''
    return str(datetime.datetime.fromtimestamp(x))


def yamlize(x):
    ret = yaml.dump(x, default_flow_style=False)
    return ret


markdown = mistune.Markdown(hard_wrap=True)


def make_hierarchies(statuses):

    def group(l):
        pipelines = list(filter(lambda x: len(x['id']) == 1, l))
        children_ = list(filter(lambda x: len(x['id']) > 1, l))
        groups_ = {}
        for child in children_:
            child_key = child['id'].pop(0)
            groups_.setdefault(child_key, []).append(child)
        children_ = dict(
            (k, group(v))
            for k, v in groups_.items()
        )
        for p in pipelines:
            p['id'] = p['id'][0]
        return {
            'pipelines': pipelines,
            'children': children_
        }

    def flatten(children_):
        for k, v in children_.items():
            v['children'] = flatten(v['children'])
            child_keys = list(v['children'].keys())
            if len(child_keys) == 1 and len(v['pipelines']) == 0:
                child_key = child_keys[0]
                children_['/'.join([k, child_key])] = v['children'][child_key]
                del children_[k]
        return children_

    statuses = [
       {
            'id': st['id'].split('/'),
            'title': st.get('title'),
            'stats': st.get('stats'),
            'slug': st.get('slug')
       }
       for st in statuses
    ]
    groups = group(statuses)
    children = groups.get('children', {})
    groups['children'] = flatten(children)

    return groups


blueprint = Blueprint('dpp', 'dpp')


@blueprint.route("")
def main():
    all_pipeline_ids = sorted(status.all_pipeline_ids())
    statuses = []
    for pipeline_id in all_pipeline_ids:
        pipeline_status = status.get(pipeline_id)
        ex = pipeline_status.get_last_execution()
        success_ex = pipeline_status.get_last_successful_execution()
        pipeline_obj = {
            'id': pipeline_id.lstrip('./'),
            'title': pipeline_status.pipeline_details.get('title'),
            'stats': user_facing_stats(ex.stats) if ex else None,
            'slug': slugify.slugify(pipeline_id),
            'trigger': ex.trigger if ex else None,
            'error_log': pipeline_status.errors(),
            'state': pipeline_status.state(),
            'pipeline': pipeline_status.pipeline_details,
            'message': pipeline_status.state().capitalize(),
            'dirty': pipeline_status.dirty(),
            'runnable': pipeline_status.runnable(),
            'class': {'INIT': 'primary',
                      'QUEUED': 'primary',
                      'INVALID': 'danger',
                      'RUNNING': 'warning',
                      'SUCCEEDED': 'success',
                      'FAILED': 'danger'
                      }[pipeline_status.state()],
            'ended': datestr(ex.finish_time) if ex else None,
            'started': datestr(ex.start_time) if ex else None,
            'last_success': datestr(success_ex.finish_time) if success_ex else None,
        }
        statuses.append(pipeline_obj)

    def state_and_not_dirty(state, p):
        return p.get('state') == state and not p.get('dirty')

    def state_or_dirty(state, p):
        return p.get('state') == state or p.get('dirty')

    categories = [
        ['ALL', 'All Pipelines', lambda _, __: True],
        ['INVALID', "Can't start", lambda _, p: not p['runnable']],
        ['QUEUED', 'Waiting to run', lambda state, p: p['state'] == state],
        ['RUNNING', 'Running', state_and_not_dirty],
        ['FAILED', 'Failed Execution', state_and_not_dirty],
        ['SUCCEEDED', 'Successful Execution', state_and_not_dirty],
    ]
    for item in categories:
        item.append([p for p in deepcopy(statuses)
                     if item[2](item[0], p)])
        item.append(len(item[-1]))
        item.append(make_hierarchies(item[-2]))
    return render_template('dashboard.html',
                           categories=categories,
                           yamlize=yamlize,
                           markdown=markdown)


@blueprint.route("api/refresh")
def refresh():
    execute_update_pipelines()
    return jsonify({'ok': True})


@blueprint.route("api/raw/status")
def pipeline_raw_api_status():
    pipelines = sorted(status.all_statuses(), key=lambda x: x.get('id'))
    for pipeline in pipelines:
        # can get the full details from api/raw/<path:pipeline_id>
        for attr in ["pipeline", "reason", "error_log"]:
            if attr in pipeline:
                del pipeline[attr]
    return jsonify(pipelines)


@blueprint.route("api/raw/<path:pipeline_id>")
def pipeline_raw_api(pipeline_id):
    if not pipeline_id.startswith('./'):
        pipeline_id = './' + pipeline_id
    pipeline_status = status.get(pipeline_id)
    if not pipeline_status.pipeline_details:
        abort(404)
    last_execution = pipeline_status.get_last_execution()
    last_successful_execution = pipeline_status.get_last_successful_execution()
    ret = {
        "id": pipeline_id,
        "cache_hash": pipeline_status.cache_hash,
        "dirty": pipeline_status.dirty(),

        "queued": last_execution.queue_time if last_execution else None,
        "started": last_execution.start_time if last_execution else None,
        "ended": last_execution.finish_time if last_execution else None,
        "reason": last_execution.log if last_execution else None,
        "error_log": pipeline_status.errors(),
        "stats": last_execution.stats if last_execution else None,
        "success": last_execution.success if last_execution else None,
        "last_success": last_successful_execution.finish_time if last_successful_execution else None,
        "trigger": last_execution.trigger if last_execution else None,

        "pipeline": pipeline_status.pipeline_details,
        "source": pipeline_status.source_spec,
        "message": pipeline_status.state().capitalize(),
        "state": pipeline_status.state(),
    }

    return jsonify(ret)


@blueprint.route("api/<field>/<path:pipeline_id>")
def pipeline_api(field, pipeline_id):

    if not pipeline_id.startswith('./'):
        pipeline_id = './' + pipeline_id
    pipeline_status = status.get(pipeline_id)
    if not pipeline_status.pipeline_details:
        abort(404)

    ret = None
    if field == 'pipeline':
        ret = pipeline_status.pipeline_details
        ret = yamlize(ret)
    elif field == 'source':
        ret = pipeline_status.source_spec
        ret = yamlize(ret)
    elif field == 'log':
        ex = pipeline_status.get_last_execution()
        ret = ex.log if ex else ''
    else:
        abort(400)

    ret = ret.split('\n')
    ret = {'text': ret}
    return jsonify(ret)


@blueprint.route("badge/<path:pipeline_id>")
def badge(pipeline_id):
    if not pipeline_id.startswith('./'):
        pipeline_id = './' + pipeline_id
    pipeline_status = status.get_status(pipeline_id)
    if pipeline_status is None:
        abort(404)
    status_text = pipeline_status.get('message')
    success = pipeline_status.get('success')
    if success is True:
        record_count = pipeline_status.get('stats', {}).get('total_row_count')
        if record_count is not None:
            status_text += ' (%d records)' % record_count
        status_color = 'brightgreen'
    elif success is False:
        status_color = 'red'
    else:
        status_color = 'lightgray'
    return redirect('https://img.shields.io/badge/{}-{}-{}.svg'.format(
        'pipeline', status_text, status_color
    ))


app = Flask(__name__)
app.config['JSONIFY_PRETTYPRINT_REGULAR'] = True
CORS(app)

url_prefix = os.environ.get('DPP_BASE_PATH', '/')
if not url_prefix.endswith('/'):
    url_prefix += '/'
logging.info('Serving on path %s', url_prefix)

app.register_blueprint(blueprint, url_prefix=url_prefix)



import copy
import os
import collections

from datapackage_pipelines.wrapper import ingest, spew
from datapackage_pipelines.utilities.kvstore import KVStore
from datapackage_pipelines.utilities.resources import PROP_STREAMING

db = KVStore()


class KeyCalc(object):

    def __init__(self, key_spec):
        if isinstance(key_spec, list):
            key_spec = ':'.join('{%s}' % key for key in key_spec)
        self.key_spec = key_spec

    def __call__(self, row):
        return self.key_spec.format(**row)


def identity(x):
    return x


Aggregator = collections.namedtuple('Aggregator',
                                    ['func', 'finaliser', 'dataType', 'copyProperties'])
AGGREGATORS = {
    'sum': Aggregator(lambda curr, new:
                      new + curr if curr is not None else new,
                      identity,
                      None,
                      False),
    'avg': Aggregator(lambda curr, new:
                      (curr[0] + 1, new + curr[1])
                      if curr is not None
                      else (1, new),
                      lambda value: value[1] / value[0],
                      None,
                      False),
    'max': Aggregator(lambda curr, new:
                      max(new, curr) if curr is not None else new,
                      identity,
                      None,
                      False),
    'min': Aggregator(lambda curr, new:
                      min(new, curr) if curr is not None else new,
                      identity,
                      None,
                      False),
    'first': Aggregator(lambda curr, new:
                        curr if curr is not None else new,
                        identity,
                        None,
                        True),
    'last': Aggregator(lambda curr, new: new,
                       identity,
                       None,
                       True),
    'count': Aggregator(lambda curr, new:
                        curr+1 if curr is not None else 1,
                        identity,
                        'integer',
                        False),
    'any': Aggregator(lambda curr, new: new,
                      identity,
                      None,
                      True),
    'set': Aggregator(lambda curr, new:
                      curr.union({new}) if curr is not None else {new},
                      lambda value: list(value) if value is not None else [],
                      'array',
                      False),
    'array': Aggregator(lambda curr, new:
                        curr + [new] if curr is not None else [new],
                        identity,
                        'array',
                        False),
}

parameters, datapackage, resource_iterator = ingest()

source = parameters['source']
source_name = source['name']
source_key = KeyCalc(source['key'])
source_delete = source.get('delete', False)

target = parameters['target']
target_name = target['name']
if target['key'] is not None:
    target_key = KeyCalc(target['key'])
else:
    target_key = None
deduplication = target_key is None


def fix_fields(fields_):
    for field in sorted(fields_.keys()):
        spec = fields_[field]
        if spec is None:
            fields_[field] = spec = {}
        if 'name' not in spec:
            spec['name'] = field
        if 'aggregate' not in spec:
            spec['aggregate'] = 'any'
    return fields_


fields = fix_fields(parameters['fields'])
full = parameters.get('full', True)


def indexer(resource):
    for row in resource:
        key = source_key(row)
        try:
            current = db[key]
        except KeyError:
            current = {}
        for field, spec in fields.items():
            name = spec['name']
            curr = current.get(field)
            agg = spec['aggregate']
            if agg != 'count':
                new = row.get(name)
            else:
                new = ''
            if new is not None:
                current[field] = AGGREGATORS[agg].func(curr, new)
        db[key] = current
        yield row


def process_target(resource):
    if deduplication:
        # just empty the iterable
        collections.deque(indexer(resource), maxlen=0)
        for key in db.keys():
            row = dict(
                (f, None) for f in fields.keys()
            )
            row.update(dict(
                (k, AGGREGATORS[fields[k]['aggregate']].finaliser(v))
                for k, v in db.db.get(key).items()
            ))
            yield row
    else:
        for row in resource:
            key = target_key(row)
            try:
                extra = db[key]
                extra = dict(
                    (k, AGGREGATORS[fields[k]['aggregate']].finaliser(v))
                    for k, v in extra.items()
                )
            except KeyError:
                if not full:
                    continue
                extra = dict(
                    (k, row.get(k))
                    for k in fields.keys()
                )
            row.update(extra)
            yield row


def new_resource_iterator(resource_iterator_):
    has_index = False
    for resource in resource_iterator_:
        name = resource.spec['name']
        if name == source_name:
            has_index = True
            if source_delete:
                # just empty the iterable
                collections.deque(indexer(resource), maxlen=0)
            else:
                yield indexer(resource)
            if deduplication:
                yield process_target(resource)
        elif name == target_name:
            assert has_index
            yield process_target(resource)
        else:
            yield resource


def process_target_resource(source_spec, resource):
    target_fields = \
        resource.setdefault('schema', {}).setdefault('fields', [])
    added_fields = sorted(fields.keys())
    for field in added_fields:
        spec = fields[field]
        agg = spec['aggregate']
        data_type = AGGREGATORS[agg].dataType
        copy_properties = AGGREGATORS[agg].copyProperties
        to_copy = {}
        if data_type is None:
            source_field = \
                next(filter(lambda f, spec_=spec:
                            f['name'] == spec_['name'],
                            source_spec['schema']['fields']))
            if copy_properties:
                to_copy = copy.deepcopy(source_field)
            data_type = source_field['type']
        try:
            existing_field = next(iter(filter(
                lambda f: f['name'] == field,
                target_fields)))
            assert existing_field['type'] == data_type, \
                'Reusing %s but with different data types: %s != %s' % (field, existing_field['type'], data_type)
        except StopIteration:
            to_copy.update({
                'name': field,
                'type': data_type
            })
            target_fields.append(to_copy)
    return resource


def process_datapackage(datapackage_):

    new_resources = []
    source_spec = None

    for resource in datapackage_['resources']:

        if resource['name'] == source_name:
            source_spec = resource
            if not source_delete:
                new_resources.append(resource)
            if deduplication:
                resource = process_target_resource(
                    source_spec,
                    {
                        'name': target_name,
                        PROP_STREAMING: True,
                        'path': os.path.join('data', target_name + '.csv')
                    })
                new_resources.append(resource)

        elif resource['name'] == target_name:
            assert isinstance(source_spec, dict), \
                "Source resource must appear before target resource"
            resource = process_target_resource(source_spec, resource)
            new_resources.append(resource)

        else:
            new_resources.append(resource)

    datapackage_['resources'] = new_resources
    return datapackage_


spew(process_datapackage(datapackage),
     new_resource_iterator(resource_iterator))
